import unittest

import bitarray

from pyais import encode_dict, encode_msg
from pyais.constants import InlandLoadedType, NavigationStatus
from pyais.decode import decode
from pyais.encode import data_to_payload, get_ais_type
from pyais.exceptions import UnknownPartNoException
from pyais.messages import MessageType1, MessageType26BroadcastUnstructured, MessageType26AddressedUnstructured, \
    MessageType26BroadcastStructured, MessageType26AddressedStructured, MessageType25BroadcastUnstructured, \
    MessageType25AddressedUnstructured, MessageType25BroadcastStructured, MessageType25AddressedStructured, \
    MessageType24PartB, MessageType24PartA, MessageType22Broadcast, MessageType22Addressed, MessageType27, \
    MessageType23, MessageType21, MessageType20, MessageType19, MessageType18, MessageType17, MessageType16, \
    MessageType15, MessageType4, MessageType5, MessageType6, MessageType7, MessageType8Default, MessageType2, MessageType3, \
    MSG_CLASS
from pyais.util import decode_bin_as_ascii6, decode_into_bit_array, str_to_bin, int_to_bin, to_six_bit, encode_ascii_6, \
    int_to_bytes, bits2bytes


def test_widths():
    tot_width = sum(field.metadata['width'] for field in MessageType1.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType4.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType5.fields())
    assert tot_width == 424

    tot_width = sum(field.metadata['width'] for field in MessageType6.fields())
    assert tot_width == 1008

    tot_width = sum(field.metadata['width'] for field in MessageType7.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType8Default.fields())
    assert tot_width == 1008

    tot_width = sum(field.metadata['width'] for field in MessageType15.fields())
    assert tot_width == 160

    tot_width = sum(field.metadata['width'] for field in MessageType15.fields())
    assert tot_width == 160

    tot_width = sum(field.metadata['width'] for field in MessageType16.fields())
    assert tot_width == 144

    tot_width = sum(field.metadata['width'] for field in MessageType17.fields())
    assert tot_width == 816

    tot_width = sum(field.metadata['width'] for field in MessageType18.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType19.fields())
    assert tot_width == 312

    tot_width = sum(field.metadata['width'] for field in MessageType20.fields())
    assert tot_width == 160

    tot_width = sum(field.metadata['width'] for field in MessageType21.fields())
    assert tot_width == 360

    tot_width = sum(field.metadata['width'] for field in MessageType23.fields())
    assert tot_width == 160

    tot_width = sum(field.metadata['width'] for field in MessageType27.fields())
    assert tot_width == 96


def test_variable_message_length_width():
    # 22
    tot_width = sum(field.metadata['width'] for field in MessageType22Addressed.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType22Broadcast.fields())
    assert tot_width == 168

    # 24
    tot_width = sum(field.metadata['width'] for field in MessageType24PartA.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType24PartB.fields())
    assert tot_width == 168

    # 25
    tot_width = sum(field.metadata['width'] for field in MessageType25AddressedStructured.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType25BroadcastStructured.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType25AddressedUnstructured.fields())
    assert tot_width == 168

    tot_width = sum(field.metadata['width'] for field in MessageType25BroadcastUnstructured.fields())
    assert tot_width == 168

    # 26
    classes = [MessageType26AddressedStructured, MessageType26BroadcastStructured,
               MessageType26AddressedUnstructured, MessageType26BroadcastUnstructured]

    for cls in classes:
        tot_width = sum(field.metadata['width'] for field in cls.fields())
        assert tot_width == 1064


def test_encode_msg_table():
    """
    Make sure that each message number as the correct Message class associated
    """
    for k, v in list(MSG_CLASS.items())[1:]:
        if k < 10:
            assert str(k) == v.__name__[-1:]
        else:
            assert str(k) == v.__name__[-2:]


def test_invalid_talker_id():
    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_dict({'mmsi': 123456}, talker_id="AIDDD")

    assert str(err.exception) == "talker_id must be any of ['AIVDM', 'AIVDO']"

    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_dict({'mmsi': 123456}, talker_id=None)

    assert str(err.exception) == "talker_id must be any of ['AIVDM', 'AIVDO']"


def test_encode_payload_invalid_talker_id():
    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_msg({'mmsi': 123456}, talker_id="AIDDD")

    assert str(err.exception) == "talker_id must be any of ['AIVDM', 'AIVDO']"

    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_msg({'mmsi': 123456}, talker_id=None)

    assert str(err.exception) == "talker_id must be any of ['AIVDM', 'AIVDO']"


def test_invalid_radio_channel():
    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_dict({'mmsi': 123456}, radio_channel="C")

    assert str(err.exception) == "radio_channel must be any of ['A', 'B']"

    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_dict({'mmsi': 123456}, radio_channel=None)

    assert str(err.exception) == "radio_channel must be any of ['A', 'B']"


def test_encode_payload_error_radio():
    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_msg({'mmsi': 123456}, radio_channel="C")

    assert str(err.exception) == "radio_channel must be any of ['A', 'B']"

    with unittest.TestCase().assertRaises(ValueError) as err:
        encode_msg({'mmsi': 123456}, radio_channel=None)

    assert str(err.exception) == "radio_channel must be any of ['A', 'B']"


def test_data_to_payload():
    assert data_to_payload(0, {'mmsi': 123}).__class__ == MessageType1
    assert data_to_payload(1, {'mmsi': 123}).__class__ == MessageType1
    assert data_to_payload(2, {'mmsi': 123}).__class__ == MessageType2
    assert data_to_payload(3, {'mmsi': 123}).__class__ == MessageType3
    assert data_to_payload(4, {'mmsi': 123}).__class__ == MessageType4
    assert data_to_payload(5, {'mmsi': 123}).__class__ == MessageType5
    assert data_to_payload(6, {'mmsi': 123, 'dest_mmsi': 1234}).__class__ == MessageType6
    assert data_to_payload(7, {'mmsi': 123}).__class__ == MessageType7
    assert data_to_payload(8, {'mmsi': 123}).__class__ == MessageType8Default

    with unittest.TestCase().assertRaises(ValueError):
        data_to_payload(28, {'mmsi': 123})


def test_get_ais_type():
    ais_type = get_ais_type({'type': 1})
    assert ais_type == 1

    ais_type = get_ais_type({'type': '1'})
    assert ais_type == 1

    ais_type = get_ais_type({'msg_type': 1})
    assert ais_type == 1

    ais_type = get_ais_type({'msg_type': '1'})
    assert ais_type == 1

    with unittest.TestCase().assertRaises(ValueError) as err:
        get_ais_type({})
    assert str(err.exception) == "Missing or invalid AIS type. Must be a number."

    with unittest.TestCase().assertRaises(ValueError) as err:
        get_ais_type({'typee': 1})
    assert str(err.exception) == "Missing or invalid AIS type. Must be a number."


def test_str_to_bin():
    # Test that Hello is correctly converted
    string = str_to_bin("Hello", 5 * 6).to01()
    assert string == "001000000101001100001100001111"
    assert len(string) == 30

    # Test that at most width characters are encoded
    string = str_to_bin("Hello World", 5 * 6).to01()
    assert string == "001000000101001100001100001111"
    assert len(string) == 30

    # By default, no trailing spaces should be added
    string = str_to_bin("Hello", 96).to01()
    assert string == "001000000101001100001100001111"
    assert len(string) == 30

    # But trailing spaces can be added
    string = str_to_bin("Hello", 96, trailing_spaces=True).to01()
    assert string == "001000000101001100001100001111000000000000000000000000000000000000000000000000000000000000000000"
    assert len(string) == 96


def test_int_to_bin():
    num = int_to_bin(0, 10).to01()
    assert num == "0000000000"
    assert len(num) == 10

    num = int_to_bin(6, 10).to01()
    assert num == "0000000110"
    assert len(num) == 10

    num = int_to_bin(128, 7).to01()
    assert num == "1111111"
    assert len(num) == 7

    num = int_to_bin(255, 8).to01()
    assert num == "11111111"
    assert len(num) == 8


def test_decode_encode():
    """Create each message with default values and test that it can be decoded again"""
    mmsi = 123
    for typ in MSG_CLASS.keys():
        encoded = encode_dict({'mmsi': mmsi, 'dest_mmsi': 656634123, 'type': typ})
        decoded = decode(*encoded).asdict()

        assert decoded['mmsi'] == 123
        if 'dest_mmsi' in decoded:
            assert decoded['dest_mmsi'] == 656634123


def test_encode_type_27():
    data = {
        'accuracy': 0,
        'course': 167,
        'gnss': 0,
        'lat': 4.84,
        'lon': 137.02333333333334,
        'mmsi': '206914217',
        'raim': 0,
        'repeat': 0,
        'speed': 57,
        'status': 2,
        'type': 27
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,K35E2b@U19PFdLbL,0*71"


def test_encode_type_26():
    data = {
        'addressed': 0,
        'data': b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc0",
        'mmsi': '016777280',
        'radio': 647746,
        'repeat': 0,
        'structured': 0,
        'type': 26
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,J0@00@3wwwwwwwwwwww0WR@P,4*4C"


def test_encode_type_25_b():
    data = {
        'addressed': 1,
        'data': b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc0",
        'dest_mmsi': '134218384',
        'mmsi': '440006460',
        'repeat': 0,
        'structured': 1,
        'app_id': 45,
        'type': 25
    }
    encoded = encode_dict(data)
    assert encoded[0] == '!AIVDO,1,1,,A,I6SWo?<P00a00;Owwwwwwwwwwwt0,2*47'


def test_encode_type_25_a():
    data = {
        'addressed': 1,
        'data': b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xc0",
        'dest_mmsi': '134218384',
        'mmsi': '440006460',
        'repeat': 0,
        'structured': 0,
        'type': 25
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,I6SWo?8P00a3wwwwwwwwwwwwwwt0,2*73"


def test_encode_type_24_partno_invalid():
    # Should not raise an error
    encode_dict({'mmsi': 123, 'partno': 1, 'type': 24})

    with unittest.TestCase().assertRaises(UnknownPartNoException):
        encode_dict({'mmsi': 123, 'partno': 2, 'type': 24})

    with unittest.TestCase().assertRaises(UnknownPartNoException):
        encode_dict({'mmsi': 123, 'partno': 3, 'type': 24})


def test_encode_type_24_a():
    data = {
        'type': 24,
        'mmsi': '338091445',
        'partno': 0,
        'shipname': "HMS FooBar",
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,H52KMe@Pm>0Htt85800000000000,0*36"


def test_encode_type_24_b():
    data = {
        'callsign': '',
        'mmsi': '338091445',
        'model': 12,
        'mothership_mmsi': '000000000',
        'partno': 1,
        'repeat': 0,
        'serial': 199729,
        'ship_type': 37,  # PleasureCraft
        'to_bow': 0,
        'to_port': 0,
        'to_starboard': 0,
        'to_stern': 0,
        'type': 24,
        'vendorid': 'FEC'
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,H52KMeDU653hhhi0000000000000,0*18"


def test_encode_type_23():
    data = {
        'interval': 9,
        'mmsi': '002268120',
        'ne_lat': 3064.2000000000003,
        'ne_lon': 157.8,
        'quiet': 0,
        'repeat': 0,
        'shiptype': 0,
        'station_type': 6,
        'sw_lat': 3040.8,
        'sw_lon': 109.60000000000001,
        'txrx': 0,
        'type': 23
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,G02:Kn01R`sn@291nj600000900,2*13"


def test_encode_type_22_b():
    data = {
        'addressed': 1,
        'band_a': 0,
        'band_b': 0,
        'channel_a': 3584,
        'channel_b': 8,
        'dest1': '028144881',
        'dest2': '268435519',
        'mmsi': '017419965',
        'power': 1,
        'repeat': 0,
        'txrx': 1,
        'type': 22,
        'zonesize': 4
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,F0@W>gCP00PH=JrN84000?hB0000,0*75"


def test_encode_type_22_a():
    data = {
        'addressed': 0,
        'band_a': 0,
        'band_b': 0,
        'channel_a': 2087,
        'channel_b': 2088,
        'mmsi': '003160107',
        'ne_lat': 3300.0,
        'ne_lon': -7710.0,
        'power': 0,
        'repeat': 0,
        'sw_lat': 3210.0,
        'sw_lon': -8020.0,
        'txrx': 0,
        'type': 22,
        'zonesize': 2
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,F030p:j2N2P5aJR0r;6f3rj10000,0*10"


def test_encode_type_21():
    data = {
        'accuracy': 1,
        'aid_type': 1,  # Reference point
        'assigned': 0,
        'epfd': 1,  # GPS
        'lat': 48.65457,
        'lon': -123.429155,
        'mmsi': '316021442',
        'name': 'DFO2',
        'name_extension': '',
        'off_position': 1,
        'raim': 1,
        'reserved_1': 0,
        'repeat': 0,
        'second': 18,
        'to_bow': 0,
        'to_port': 0,
        'to_starboard': 0,
        'to_stern': 0,
        'type': 21,
        'virtual_aid': 0
    }

    encoded = encode_dict(data)
    # Validated using: http://ais.tbsalling.dk/
    assert encoded[0] == "!AIVDO,1,1,,A,E4eHJhPR37q0000000000000000KUOSc=rq4h00000a@2000000000000000,4*39"

    data['reserved_1'] = 255
    encoded = encode_dict(data)
    # Validated using: http://ais.tbsalling.dk/
    assert encoded[0] == "!AIVDO,1,1,,A,E4eHJhPR37q0000000000000000KUOSc=rq4h00000aOv000000000000000,4*72"


def test_encode_type_20():
    data = {
        'increment1': 750,
        'increment2': 0,
        'increment3': 0,
        'increment4': 0,
        'mmsi': '002243302',
        'number1': 5,
        'number2': 0,
        'number3': 0,
        'number4': 0,
        'offset1': 200,
        'offset2': 0,
        'offset3': 0,
        'offset4': 0,
        'repeat': 0,
        'timeout1': 7,
        'timeout2': 0,
        'timeout3': 0,
        'timeout4': 0,
        'type': 20
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,D028rqP<QNfp000000000000000,2*0E"


def test_encode_type_19():
    data = {
        'accuracy': 0,
        'assigned': 0,
        'course': 335.90000000000003,
        'dte': 0,
        'epfd': 1,  # GPS
        'heading': 511,
        'lat': 29.543695,
        'lon': -88.81039166666666,
        'mmsi': '367059850',
        'raim': 0,
        'reserved_2': 4,
        'repeat': 0,
        'second': 46,
        'shipname': 'CAPT.J.RIMES',
        'ship_type': 70,  # CARGO
        'speed': 5.5,
        'to_bow': 5,
        'to_port': 4,
        'to_starboard': 4,
        'to_stern': 21,
        'type': 19
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,C5N3SRP0=nJGEBT>NhWAwwo862PaLELTBJ:V00000000S0D:R220,0*25"

    data['ship_type'] = 255
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,C5N3SRP0=nJGEBT>NhWAwwo862PaLELTBJ:V0000000000D:R220,0*46"


def test_encode_type_18_with_speed_and_course():
    data = {
        'accuracy': 0,
        'assigned': 0,
        'band': 1,
        'course': 10.1,
        'cs': 1,
        'display': 0,
        'dsc': 1,
        'heading': 511,
        'lat': 37.785035,
        'lon': -122.26732,
        'mmsi': '367430530',
        'msg22': 1,
        'radio': 917510,
        'raim': 0,
        'regional': 0,
        'repeat': 0,
        'second': 55,
        'speed': 67.85,
        'type': 18
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,B5NJ;PP2aUl4ot5Isbl6GwsUkP06,0*35"


def test_encode_type_18():
    data = {
        'accuracy': 0,
        'assigned': 0,
        'band': 1,
        'course': 0.0,
        'cs': 1,
        'display': 0,
        'dsc': 1,
        'heading': 511,
        'lat': 37.785035,
        'lon': -122.26732,
        'mmsi': '367430530',
        'msg22': 1,
        'radio': 917510,
        'raim': 0,
        'regional': 0,
        'repeat': 0,
        'second': 55,
        'speed': 0.0,
        'type': 18
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,B5NJ;PP005l4ot5Isbl03wsUkP06,0*74"

    data['heading'] = 123
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,B5NJ;PP005l4ot5Isbl00usUkP06,0*75"


def test_encode_type_17_b():
    data = {
        'data': bits2bytes('00000011101011001011110001000110001111011111111111000100'),
        'lat': 2058.2,
        'lon': 8029.2,
        'mmsi': '004310602',
        'repeat': 0,
        'type': 17
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,A0476BQ>J@`<h0>dg4Huwt@,2*36"


def test_encode_type_17_a():
    data = {
        'data': bits2bytes('00000011101011001011110001000110001111011111111111000100'),
        'lat': 3599.2,
        'lon': 1747.8,
        'mmsi': '002734450',
        'repeat': 0,
        'type': 17
    }

    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,A02VqLPA4I6C00>dg4Huwt@,2*60"


def test_encode_type_16():
    data = {
        'increment1': 0,
        'increment2': 0,
        'mmsi': '002053501',
        'mmsi1': '224251000',
        'mmsi2': '000000000',
        'offset1': 200,
        'offset2': 0,
        'repeat': 0,
        'type': 16
    }
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,@01uEO@mMk7P<P0000000000,0*1A"


def test_encode_type_15_a():
    data = {
        'mmsi': '003669720',
        'mmsi1': '367014320',
        'mmsi2': '000000000',
        'offset1_1': 516,
        'offset1_2': 617,
        'offset2_1': 0,
        'repeat': 3,
        'type': 15,
        'type1_1': 3,
        'type1_2': 5,
        'type2_1': 0
    }
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,?h3Ovn1GP<K0<P@59a000000000,2*05"


def test_encode_type_15():
    data = {
        'mmsi': '368578000',
        'mmsi1': '000005158',
        'mmsi2': '000000000',
        'offset1_1': 0,
        'offset1_2': 0,
        'offset2_1': 0,
        'repeat': 0,
        'type': 15,
        'type1_1': 5,
        'type1_2': 0,
        'type2_1': 0
    }
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,?5OP=l00052HD00000000000000,2*59"


def test_encode_type_14():
    data = {'mmsi': '351809000', 'repeat': 0, 'text': 'RCVD YR TEST MSG', 'type': 14}
    actual = encode_dict(data)
    expected = ['!AIVDO,1,1,,A,>5?Per18=HB1U:1@E=B0m<L,2*53']

    assert expected == actual


def test_encode_type_13():
    data = {
        'mmsi': '211378120',
        'mmsi1': '211217560',
        'mmsi2': '000000000',
        'mmsi3': '000000000',
        'mmsi4': '000000000',
        'mmsiseq1': 2,
        'mmsiseq2': 0,
        'mmsiseq3': 0,
        'mmsiseq4': 0,
        'repeat': 0,
        'type': 13
    }
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,739UOj0jFs9R0000000000000000,0*6D"


def test_encode_type_12_with_some_text():
    data = {
        'dest_mmsi': '271002111',
        'mmsi': '271002099',
        'repeat': 0,
        'retransmit': 1,
        'seqno': 0,
        'text': 'MSG FROM 271002099',
        'type': 12
    }
    actual = encode_dict(data)
    expected = ['!AIVDO,1,1,,A,<42Lati0W:Ov=C7P6B?=Pjoihhjhqq,0*1B', ]

    assert expected == actual


def test_encode_type_12_with_no_text():
    data = {
        'dest_mmsi': '271002111',
        'mmsi': '271002099',
        'repeat': 0,
        'retransmit': 1,
        'seqno': 0,
        'text': None,
        'type': 12
    }
    actual = encode_dict(data)
    expected = ['!AIVDO,1,1,,A,<42Lati0W:Ov,0*4A', ]

    assert expected == actual


def test_encode_type_12_with_empty_text():
    data = {
        'dest_mmsi': '271002111',
        'mmsi': '271002099',
        'repeat': 0,
        'retransmit': 1,
        'seqno': 0,
        'text': '',
        'type': 12
    }
    actual = encode_dict(data)
    expected = ['!AIVDO,1,1,,A,<42Lati0W:Ov,0*4A', ]

    assert expected == actual


def test_encode_type_12_with_max_length_text():
    data = {
        'dest_mmsi': '271002111',
        'mmsi': '271002099',
        'repeat': 0,
        'retransmit': 1,
        'seqno': 0,
        'text': 156 * 'Q',
        'type': 12
    }
    actual = encode_dict(data)
    expected = [
        '!AIVDO,3,1,0,A,<42Lati0W:OvAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,0*78',
        '!AIVDO,3,2,0,A,AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,0*15',
        '!AIVDO,3,3,0,A,AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,0*14'
    ]

    assert expected == actual


def test_encode_type_12_with_longer_than_max_length_length_text():
    data = {
        'dest_mmsi': '271002111',
        'mmsi': '271002099',
        'repeat': 0,
        'retransmit': 1,
        'seqno': 0,
        'text': 160 * 'Q',
        'type': 12
    }
    actual = encode_dict(data)
    expected = [
        '!AIVDO,3,1,0,A,<42Lati0W:OvAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,0*78',
        '!AIVDO,3,2,0,A,AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,0*15',
        '!AIVDO,3,3,0,A,AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA,0*14'
    ]

    assert expected == actual


def test_encode_type_11():
    data = {
        'accuracy': 1,
        'day': 22,
        'epfd': 1,
        'hour': 2,
        'lat': 28.409117,
        'lon': -94.40768,
        'minute': 22,
        'mmsi': '304137000',
        'month': 5,
        'radio': 0,
        'raim': 0,
        'repeat': 0,
        'second': 40,
        'type': 11,
        'year': 2009
    }
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,44R33:1uUK2F`q?mP0@@GoQ00000,0*08"


def test_encode_type_10():
    data = {'dest_mmsi': '366972000', 'mmsi': '440882000', 'repeat': 0, 'type': 10}
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,:6TMCD1GOS60,0*5A"


def test_encode_type_9():
    data = {
        'accuracy': 0,
        'alt': 303,
        'assigned': 0,
        'course': 154.5,
        'dte': 1,
        'lat': 58.144,
        'lon': -6.2788434,
        'mmsi': '111232511',
        'radio': 33392,
        'raim': 0,
        'repeat': 0,
        'second': 15,
        'speed': 42,
        'type': 9
    }
    encoded = encode_dict(data)
    assert encoded[0] == "!AIVDO,1,1,,A,91b55wi;hbOS@OdQAC062Ch2089h,0*31"


def test_encode_type_8():
    data = {
        'dac': 366,
        'data': bits2bytes('00000011101011001011110001000110001111011111111111000100'),
        'fid': 56,
        'mmsi': '366999712',
        'repeat': 0,
        'type': 8
    }
    encoded = encode_dict(data, radio_channel="B", talker_id="AIVDM")
    assert encoded[0] == "!AIVDM,1,1,,B,85Mwp`1Kf0>dg4Huwt@,2*5B"


def test_encode_type_8_inland():
    data = {
        'msg_type': 8,
        'repeat': 0,
        'mmsi': 366053209,
        'spare_1': b'\x00',
        'dac': 200,
        'fid': 10,
        'vin': 'T4]V\\6IG',
        'length': 180.6,
        'beam': 42.0,
        'shiptype': 10444,
        'hazard': 4,
        'draught': 9.47,
        'loaded': InlandLoadedType.NotAvailable,
        'spare': b'm',
        'speed_q': False,
        'course_q': True,
        'heading_q': True,
    }
    encoded = encode_dict(data, radio_channel="A", talker_id="AIVDO")
    assert encoded[0] == "!AIVDO,1,1,,A,85M67F@j2U=7EW=RAkQkBDITMV=e,0*51"


def tes_encode_type_8_multi():
    data = {
        'dac': 0,
        "data": b"\x02\x934D\x81nI;\xbd\xcd\xe5\xb7E\xed\xf1]\xc0[y\xfa#-\xcd<\x01\x05\x91\xef\x85\x92\xfbF\xed\x19t\x11\xd6\xe7\xdf\xec\x1fp\x97\x99\x83M\x8aK\xb8\x005'\x1f\xc7\x14\xeaTr\xe3o\xb8\xda\xb9\x17-FJxb\xeb5\x1aM",
        "fid": 0,
        "mmsi": 888888888,
        "type": 8
    }
    encoded = encode_dict(data)
    assert encoded == ['!AIVDO,2,1,0,A,8=?eN>0000:C=4B1KTTsgLoUelGetEo0FoWr8jo=?045TNv5Tge6sAUl4MKWo,0*5F', '!AIVDO,2,2,0,A,vhOL9NIPln:BsP0=BLOiiCbE7;SKsSJfALeATapHfdm6Tl,2*79']


def test_encode_type_7():
    data = {
        'mmsi': '002655651',
        'mmsi1': '265538450',
        'mmsi2': '000000000',
        'mmsiseq1': 0,
        'mmsiseq2': 0,
        'repeat': 0,
        'type': 7
    }
    encoded_part_1 = encode_dict(data, radio_channel="B", talker_id="AIVDM")[0]
    assert encoded_part_1 == "!AIVDM,1,1,,B,702R5`hwCjq80000000000000000,0*68"


def test_encode_type_6_bytes():
    data = {
        'dac': 669,
        'data': b'\x03\xac\xbcF=\xff\xc4',
        'dest_mmsi': '313240222',
        'fid': 11,
        'mmsi': '150834090',
        'repeat': 1,
        'retransmit': 0,
        'seqno': 3,
        'type': 6
    }
    encoded = encode_dict(data, radio_channel="B", talker_id="AIVDM")
    assert encoded[0] == "!AIVDM,1,1,,B,6B?n;be:cbapald3c;i6?Ow4,0*78"


def test_encode_type_6():
    data = {
        'dac': 669,
        'data': b"\xeb\x11\x8f\x7f\xf2",
        'dest_mmsi': 313240222,
        'fid': 11,
        'mmsi': 150834090,
        'repeat': 1,
        'retransmit': 0,
        'seqno': 3,
        'type': 6
    }
    encoded = encode_dict(data, radio_channel="B", talker_id="AIVDM")

    decoded = decode(*encoded).asdict()

    assert decoded['dac'] == data['dac']
    assert decoded['data'] == data['data']
    assert decoded['dest_mmsi'] == data['dest_mmsi']
    assert decoded['fid'] == data['fid']
    assert decoded['mmsi'] == data['mmsi']


def test_encode_type_4():
    data = {
        'accuracy': 1,
        'day': 14,
        'epfd': 7,
        'hour': 19,
        'lat': 36.88376,
        'lon': -76.3523,
        'minute': 57,
        'mmsi': '003669702',
        'month': 5,
        'radio': 67039,
        'raim': 0,
        'repeat': 0,
        'second': 39,
        'type': 4,
        'year': 2007
    }
    encoded_part_1 = encode_dict(data, radio_channel="B", talker_id="AIVDM")[0]

    assert encoded_part_1 == "!AIVDM,1,1,,B,403OviQuMGCqWrRO:HE6fD700@GO,0*3A"


def test_encode_type_5_issue_59():
    """https://github.com/M0r13n/pyais/issues/59"""
    data = {
        'ais_version': 0,
        'callsign': '3FOF8',
        'day': 15,
        'destination': 'NEW YORK',
        'draught': 12.8,
        'dte': 0,
        'epfd': 1,
        'hour': 14,
        'imo': 9134270,
        'minute': 0,
        'mmsi': '351759000',
        'month': 5,
        'repeat': 0,
        'shipname': 'EVER DIADEM',
        'shiptype': 70,
        'to_bow': 225,
        'to_port': 1,
        'to_starboard': 31,
        'to_stern': 70,
        'type': 5
    }

    actual = encode_dict(data, radio_channel="B", talker_id="AIVDM")
    expected = [
        '!AIVDM,2,1,0,B,55?MbV02;H;s<HtKP00EHE:0@T4@Dl0000000000L961O5Gf0P3QEp6ClRh0,0*75',
        '!AIVDM,2,2,0,B,00000000000,2*27'
    ]
    assert actual == expected


def test_encode_type_5():
    """
    Verified using http://ais.tbsalling.dk/.
    """
    data = {
        'ais_version': 0,
        'callsign': '3FOF8',
        'day': 15,
        'destination': 'NEW YORK',
        'draught': 12.2,
        'dte': 0,
        'epfd': 1,
        'hour': 14,
        'imo': 9134270,
        'minute': 0,
        'mmsi': '351759000',
        'month': 5,
        'repeat': 0,
        'shipname': 'EVER DIADEM',
        'shiptype': 70,
        'to_bow': 225,
        'to_port': 1,
        'to_starboard': 31,
        'to_stern': 70,
        'type': 5
    }

    actual = encode_dict(data, radio_channel="B", talker_id="AIVDM")
    expected = [
        '!AIVDM,2,1,0,B,55?MbV02;H;s<HtKP00EHE:0@T4@Dl0000000000L961O5Gf0NSQEp6ClRh0,0*0B',
        '!AIVDM,2,2,0,B,00000000000,2*27'
    ]

    assert actual == expected


def test_encode_type_5_default():
    """
    Verified using http://ais.tbsalling.dk/.
    """
    data = {'mmsi': 123456789, 'type': 5}
    actual = encode_dict(data, radio_channel="B", talker_id="AIVDM")
    expected = [
        '!AIVDM,2,1,0,B,51mg=5@00000000000000000000000000000000000000000000000000000,0*60',
        '!AIVDM,2,2,0,B,00000000000,2*27'
    ]

    assert actual == expected


def test_encode_msg_type2():
    data = {
        'accuracy': 1,
        'course': 0.0,
        'heading': 511,
        'lat': 53.542675,
        'lon': 9.979428333333333,
        'mmsi': '211512520',
        'raim': 1,
        'repeat': 2,
        'second': 34,
        'speed': 0.3,
        'turn': 0,
        'type': 2
    }
    encoded = encode_dict(data)[0]
    assert encoded == "!AIVDO,1,1,,A,1S9edj?003PecbBN`ja@0?w5R000,0*24"

    data = {
        'accuracy': 1,
        'course': 0.0,
        'heading': 511,
        'lat': 53.542675,
        'lon': 9.97942833333333,
        'mmsi': 211512520,
        'raim': 1,
        'repeat': 2,
        'second': 34,
        'speed': 0.3,
        'turn': 0,
        'type': 2,
        'status': NavigationStatus.UnderWayUsingEngine,
    }
    encoded = encode_dict(data)[0]
    assert encoded == "!AIVDO,1,1,,A,1S9edj0003PecbBN`ja@0?w5R000,0*2B"


def test_encode_msg_type_3():
    data = {
        'accuracy': 0,
        'course': 254.20000000000002,
        'heading': 217,
        'lat': 37.81065,
        'lon': -122.3343,
        'mmsi': '367581220',
        'radio': 86346,
        'raim': 0,
        'repeat': 0,
        'second': 40,
        'speed': 0.1,
        'status': 5,
        'turn': 0,
        'type': 3
    }

    encoded = encode_dict(data)[0]
    assert encoded == "!AIVDO,1,1,,A,15NSH95001G?wopE`beasVkAPE5:,0*0E"


def test_encode_type_1_default():
    """
    Verified using http://ais.tbsalling.dk/.
    """
    data = {'mmsi': 123456789, 'type': 1}
    encoded = encode_dict(data)[0]
    assert encoded == "!AIVDO,1,1,,A,11mg=5OP0000000000000001P000,0*58"


def test_encode_type_1():
    """
    Verified using http://ais.tbsalling.dk/.
    """
    data = {
        'accuracy': 0,
        'course': 219.3,
        'heading': 1,
        'lat': 37.80211833333333,
        'lon': -122.34161833333333,
        'maneuver': 0,
        'mmsi': '366053209',
        'radio': 2281,
        'raim': 0,
        'repeat': 0,
        'second': 59,
        'speed': 7.8,
        'status': 3,
        'turn': 0,
        'type': 1
    }

    encoded = encode_dict(data, radio_channel="B", talker_id="AIVDM")[0]
    assert encoded == "!AIVDM,1,1,,B,15M67FC01>G?ufbE`FepT@3n00Sa,0*53"

    encoded = encode_dict(data, radio_channel="B")[0]
    assert encoded == "!AIVDO,1,1,,B,15M67FC01>G?ufbE`FepT@3n00Sa,0*51"


def test_mmsi_too_long():
    msg = MessageType1.create(mmsi=1 << 35)
    encoded = encode_msg(msg)
    decoded = decode(encoded[0])

    assert encoded[0] == "!AIVDO,1,1,,A,1?wwwwwP0000000000000001P000,0*6C"
    assert decoded.mmsi == 1073741823


def test_lon_too_large():
    msg = MessageType1.create(mmsi="123", lon=1 << 30)
    encoded = encode_msg(msg)
    decoded = decode(encoded[0])

    assert encoded[0] == "!AIVDO,1,1,,A,10000NwP00Owwwv000000001P000,0*63"
    assert decoded.lon == -2e-06


def test_ship_name_too_lon():
    msg = MessageType5.create(mmsi="123", shipname="Titanic Titanic Titanic")
    actual = encode_msg(msg)

    expected = [
        '!AIVDO,2,1,0,A,50000Nh000000000001@U@4pT>1@U@4pT>1@U@4000000000000000000000,0*54',
        '!AIVDO,2,2,0,A,00000000000,2*26'
    ]

    assert actual == expected


def test_int_to_bytes():
    i = int_to_bytes(b'\xff\xff')
    assert i == 65535

    i = int_to_bytes(65535)
    assert i == 65535

    i = int_to_bytes(b'\x00\x00')
    assert i == 0


def test_to_six_bit():
    c = to_six_bit('a')
    assert c == '000001'

    c = to_six_bit('A')
    assert c == '000001'

    c = to_six_bit('9')
    assert c == '111001'

    with unittest.TestCase().assertRaises(ValueError):
        to_six_bit('ä')


def test_encode_ascii_6_bit():
    input_val = '001000000101001100001100001111100000010111001111010010001100000100100001'
    b = bitarray.bitarray(input_val)
    ascii6, padding = encode_ascii_6(b)

    assert ascii6 == "85<<?PG?B<4Q"
    assert padding == 0

    bit_arr = decode_into_bit_array(ascii6.encode())
    assert bit_arr.to01() == input_val
    assert decode_bin_as_ascii6(bit_arr) == "HELLO WORLD!"


def test_encode_does_not_exceed_nmea_sentence_length_limit():
    data = {
        'type': 5, 'repeat': 0, 'mmsi': '259725000', 'ais_version': 1,
        'imo': 9103128, 'callsign': 'LAXP4', 'shipname': 'STAR HANSA',
        'shiptype': 70, 'to_bow': 177, 'to_stern': 22, 'to_port': 17,
        'to_starboard': 14, 'epfd': 1, 'month': 5, 'day': 20, 'hour': 6,
        'minute': 0, 'draught': 7.4, 'destination': 'US BAL', 'dte': 0
    }

    encoded = encode_dict(data)
    encoded = [e + '\r\n' for e in encoded]

    assert len(encoded) == 2
    assert len(encoded[0]) == 82
    assert len(encoded[1]) == 33



"""
AIS filters.
A set of filters that allow filtering AIS messages based on different criteria
like attributes, empty values, message types, or geographical boundaries.
Filters can be chained using a FilterChain.
"""

import math
import socket
import typing
import pyais

# Type Aliases for readability
F = typing.TypeVar("F", typing.BinaryIO, socket.socket, None)
AIS_STREAM = pyais.Stream[F]
MESSAGE_STREAM = typing.Generator[pyais.ANY_MESSAGE, None, None]
FILTER_FUNCTION = typing.Callable[[pyais.ANY_MESSAGE], bool]
LAT_LON = typing.Tuple[float, float]  # Tuple type for latitude and longitude


def haversine(latLon1: LAT_LON, latLon2: LAT_LON) -> float:
    """
    Calculate the great circle distance between two points on the earth.

    Parameters:
    latLon1 (LAT_LON): Tuple of latitude and longitude of the first point.
    latLon2 (LAT_LON): Tuple of latitude and longitude of the second point.

    Returns:
    float: Distance between the two points in kilometers.
    """
    R = 6371.0  # Radius of the Earth in km
    lat1, lon1, lat2, lon2 = map(math.radians, [latLon1[0], latLon1[1], latLon2[0], latLon2[1]])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = math.sin(dlat / 2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2)**2
    c = 2 * math.asin(math.sqrt(a))
    return R * c


def is_in_grid(lat: float, lon: float, lat_min: float, lon_min: float, lat_max: float, lon_max: float) -> bool:
    """
    Check if a point is within a defined geographical grid.

    Parameters:
    lat, lon (float): Latitude and Longitude of the point to check.
    lat_min, lon_min, lat_max, lon_max (float): Boundaries of the grid.

    Returns:
    bool: True if the point is within the grid, False otherwise.
    """
    return lat_min <= lat <= lat_max and lon_min <= lon <= lon_max


class Filter:
    """
    Base class for all filters.
    """

    def __init__(self) -> None:
        self.next_filter: typing.Optional[Filter] = None

    def set_next(self, filter: 'Filter') -> None:
        """
        Set the next filter in the chain.

        Parameters:
        filter (Filter): The next filter to set.
        """
        self.next_filter = filter

    def filter(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Apply the filter to the data and then pass it to the next filter.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Returns:
        MESSAGE_STREAM: The filtered data stream.
        """
        data = self.filter_data(data)
        if self.next_filter:
            return self.next_filter.filter(data)
        return data

    def filter_data(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Abstract method to filter data. Should be implemented by subclasses.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Returns:
        MESSAGE_STREAM: The filtered data stream.
        """
        raise NotImplementedError("This method should be overridden by subclasses.")


class AttributeFilter(Filter):
    """
    Filter based on a user-defined function.
    """

    def __init__(self, ff: FILTER_FUNCTION) -> None:
        """
        Initialize the filter with a user-defined function.

        Parameters:
        ff (FILTER_FUNCTION): A function that takes an AIS message and returns True if the message should be kept.
        """
        super().__init__()
        self.ff = ff

    def filter_data(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Filter the data based on the user-defined function.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Yields:
        MESSAGE_STREAM: The filtered data stream.
        """
        yield from filter(self.ff, data)


class NoneFilter(Filter):
    """
    Filters messages where specified attributes are not None.
    """

    def __init__(self, *attrs: str) -> None:
        """
        Initialize the filter with attributes to check for None values.

        Parameters:
        attrs (str): Attributes that should not be None in the messages.
        """
        super().__init__()
        self.attrs = attrs

    def filter_data(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Filter the data, allowing only messages where specified attributes are not None.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Yields:
        MESSAGE_STREAM: The filtered data stream.
        """
        for msg in data:
            if all(getattr(msg, attr, None) is not None for attr in self.attrs):
                yield msg


class MessageTypeFilter(Filter):
    """
    Filters messages based on their type.
    """

    def __init__(self, *types: int) -> None:
        """
        Initialize the filter with message types to include.

        Parameters:
        types (int): Message types that should be included in the filtered data.
        """
        super().__init__()
        self.types = types

    def filter_data(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Filter the data, allowing only messages of specified types.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Yields:
        MESSAGE_STREAM: The filtered data stream.
        """
        for msg in data:
            if msg.msg_type not in self.types:
                continue
            yield msg


class DistanceFilter(Filter):
    """
    Filters messages based on distance.
    """

    def __init__(self, ref_lat_lon: LAT_LON, distance_km: float) -> None:
        """
        Initialize the filter with a reference point and distance.

        Parameters:
        ref_lat_lon (LAT_LON): Reference latitude and longitude point.
        distance_km (float): Distance threshold in kilometers.
        """
        super().__init__()
        self.ref_lat_lon = ref_lat_lon
        self.distance_km = distance_km

    def filter_data(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Filter the data based on distance from a reference point.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Yields:
        MESSAGE_STREAM: The filtered data stream.
        """
        for msg in data:
            if hasattr(msg, 'lat'):
                if haversine(self.ref_lat_lon, (msg.lat, msg.lon)) >= self.distance_km:  # type: ignore
                    continue
            yield msg


class GridFilter(Filter):
    """
    Filters messages based on a geographical grid.
    """

    def __init__(self, lat_min: float, lon_min: float, lat_max: float, lon_max: float) -> None:
        """
        Initialize the filter with grid boundaries.

        Parameters:
        lat_min, lon_min, lat_max, lon_max (float): Boundaries of the grid.
        """
        super().__init__()
        self.lat_min = lat_min
        self.lon_min = lon_min
        self.lat_max = lat_max
        self.lon_max = lon_max

    def filter_data(self, data: MESSAGE_STREAM) -> MESSAGE_STREAM:
        """
        Filter the data based on whether it falls within a specified grid.

        Parameters:
        data (MESSAGE_STREAM): The stream of data to filter.

        Yields:
        MESSAGE_STREAM: The filtered data stream.
        """
        for msg in data:
            if hasattr(msg, 'lat'):
                if not is_in_grid(msg.lat, msg.lon, self.lat_min, self.lon_min, self.lat_max, self.lon_max):  # type: ignore
                    continue
            yield msg


class FilterChain:
    """
    Chains multiple filters together.
    """

    def __init__(self, filters: typing.List[Filter]) -> None:
        """
        Initialize the filter chain with a sequence of filters.

        Parameters:
        filters (list of Filter): A list of filters to be applied in order.
        """
        if not filters:
            raise ValueError('At least one filter required')

        # Link filters together in the order provided
        for current, next in zip(filters[:-1], filters[1:]):
            current.set_next(next)

        self.filters = filters
        self.start = filters[0]

    def filter(self, stream: AIS_STREAM[F]) -> MESSAGE_STREAM:
        """
        Apply the chain of filters to the data.

        Parameters:
        stream (AIS_STREAM): The stream of data to filter.

        Yields:
        AIS_STREAM: The filtered data stream.
        """
        yield from self.start.filter(x.decode() for x in stream)



import socket
from tests.utils.timeout import time_limit
from tests.utils.skip import is_linux
import threading
import unittest
from pyais.stream import TCPConnection

MESSAGES = [
    b"!AIVDM,1,1,,B,133S0:0P00PCsJ:MECBR0gv:0D8N,0*7F",
    b"!AIVDM,1,1,,A,4h2=a@Quho;O306WMpMIK<Q00826,0*42",
    b"!AIVDM,1,1,,A,402M3b@000Htt0K0Q0R3T<700t24,0*52",
    b"!AIVDM,1,1,,A,1>qc9ww000OkfS@MMI5004R60<0B,0*31",
    b"!AIVDM,1,1,,A,13P<GAh01pwM`GPDdu>T8SDV0@2c,0*7D",
    b"!AIVDM,1,1,,A,133ma5P0000Cj9lMG484pbN60D<P,0*42",
    b"!AIVDM,1,1,,B,13aBKV5P0qPFeWJMakbGjgv820SM,0*6E",
    b"!AIVDM,1,1,,A,15Mvsu000aqSG3RF;B?A@0v4082c,0*60",
    b"!AIVDM,1,1,,A,13aI9EwP?w<tSF0l4Q@>4?wvPl6=,0*38",
    b"!AIVDM,1,1,,A,15NJIs0P0?JeI0RGBjbCCwv:282W,0*2E",
    b"!AIVDM,1,1,,A,15Mw<ePP00ISvvpA8Hi<Mwv6082J,0*45",
    b"!AIVDM,1,1,,A,15MooR0P0SJe;2>GC2pdQOv:282b,0*0C",
]


def tcp_mock_server(host, port) -> None:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, port))
    sock.listen(1)

    try:
        while True:
            # wait for a connection
            conn, _ = sock.accept()
            if conn:
                # send all at once and then close
                for msg in MESSAGES:
                    conn.send(msg + b"\r\n")
                break
    finally:
        sock.close()


class TestTCPStream(unittest.TestCase):

    def _spawn_test_server(self):
        self.server_thread = threading.Thread(target=tcp_mock_server, args=("0.0.0.0", 55555))
        self.server_thread.start()

    def test_default_buf_size(self):
        self.assertEqual(TCPConnection.BUF_SIZE, 4096)

    def test_invalid_endpoint(self):
        with self.assertRaises(ConnectionRefusedError):
            TCPConnection("0.0.0.0", 55555)

    @unittest.skipIf(not is_linux(), "Skipping because Signal is not available on non unix systems!")
    @unittest.skipIf(True, "Skip for now, because there is a Threading problem")
    def test_tcp_stream(self):
        # limit execution time to 1 second in case of potential deadlocks -> prevent test from running forever
        with time_limit(2):
            self._spawn_test_server()

            with TCPConnection("0.0.0.0", 55555) as stream:
                for i, msg in enumerate(stream):
                    assert msg.decode()
                    # make sure all messages were received
                    if i == len(MESSAGES) - 1:
                        break

        self.server_thread.join()



import unittest

from pyais import decode, encode_msg
from pyais.constants import NavigationStatus, ManeuverIndicator, ShipType, NavAid, TransmitMode, StationIntervals, \
    StationType


class TestConstants(unittest.TestCase):
    def test_that_nav_data_values_are_kept_for_all_values(self):
        # See: https://github.com/M0r13n/pyais/issues/123
        orignal = decode(b"!AIVDM,1,1,,B,15M67FC000G?ufbE`FepT@3n00Sa,0*5C")

        for i in range(15):
            orignal.status = i
            encoded = encode_msg(orignal)
            decoded = decode(encoded[0])

            self.assertEqual(decoded.status.value, i)

    def test_nav_status(self):

        self.assertEqual(NavigationStatus(3), NavigationStatus.RestrictedManoeuverability)
        self.assertEqual(NavigationStatus(17), NavigationStatus.Undefined)

    def test_maneuver_indication(self):
        self.assertEqual(ManeuverIndicator(0), ManeuverIndicator.NotAvailable)
        self.assertEqual(ManeuverIndicator(1), ManeuverIndicator.NoSpecialManeuver)
        self.assertEqual(ManeuverIndicator(2), ManeuverIndicator.SpecialManeuver)
        self.assertEqual(ManeuverIndicator(3), ManeuverIndicator.UNDEFINED)
        self.assertEqual(ManeuverIndicator(4), ManeuverIndicator.UNDEFINED)

    def test_ship_types(self):
        self.assertEqual(ShipType(25), ShipType.WIG_Reserved)
        self.assertEqual(ShipType(46), ShipType.HSC_Reserved)
        self.assertEqual(ShipType(57), ShipType.SPARE)
        self.assertEqual(ShipType(68), ShipType.Passenger_Reserved)
        self.assertEqual(ShipType(78), ShipType.Cargo_Reserved)
        self.assertEqual(ShipType(85), ShipType.Tanker_Reserved)
        self.assertEqual(ShipType(96), ShipType.OtherType_Reserved)
        self.assertEqual(ShipType(100), ShipType.NotAvailable)

    def test_navaid(self):
        self.assertEqual(NavAid.DEFAULT, 0)
        self.assertEqual(NavAid.REFERENCE_POINT, 1)
        self.assertEqual(NavAid.RACON, 2)
        self.assertEqual(NavAid.FIXED, 3)
        self.assertEqual(NavAid.EMERGENCY_WRECK_MARKING_BUOY, 4)
        self.assertEqual(NavAid.LIGHT_NO_SECTORS, 5)
        self.assertEqual(NavAid.LIGHT_WITH_SECTORS, 6)
        self.assertEqual(NavAid.LEADING_LIGHT_FRONT, 7)
        self.assertEqual(NavAid.LEADING_LIGHT_REAR, 8)
        self.assertEqual(NavAid.BEACON_CARDINAL_N, 9)
        self.assertEqual(NavAid.BEACON_CARDINAL_E, 10)
        self.assertEqual(NavAid.BEACON_CARDINAL_S, 11)
        self.assertEqual(NavAid.BEACON_CARDINAL_W, 12)
        self.assertEqual(NavAid.BEACON_PORT_HAND, 13)
        self.assertEqual(NavAid.BEACON_STARBOARD_HAND, 14)
        self.assertEqual(NavAid.BEACON_CHANNEL_PORT_HAND, 15)
        self.assertEqual(NavAid.BEACON_CHANNEL_STARBOARD_HAND, 16)
        self.assertEqual(NavAid.BEACON_ISOLATED_DANGER, 17)
        self.assertEqual(NavAid.BEACON_SAFE_WATER, 18)
        self.assertEqual(NavAid.BEACON_SPECIAL_MARK, 19)
        self.assertEqual(NavAid.CARDINAL_MARK_N, 20)
        self.assertEqual(NavAid.CARDINAL_MARK_E, 21)
        self.assertEqual(NavAid.CARDINAL_MARK_S, 22)
        self.assertEqual(NavAid.CARDINAL_MARK_W, 23)
        self.assertEqual(NavAid.PORT_HAND_MARK, 24)
        self.assertEqual(NavAid.STARBOARD_HAND_MARK, 25)
        self.assertEqual(NavAid.PREFERRED_CHANNEL_PORT_HAND, 26)
        self.assertEqual(NavAid.PREFERRED_CHANNEL_STARBOARD_HAND, 27)
        self.assertEqual(NavAid.ISOLATED_DANGER, 28)
        self.assertEqual(NavAid.SAFE_WATER, 29)
        self.assertEqual(NavAid.SPECIAL_MARK, 30)
        self.assertEqual(NavAid.LIGHT_VESSEL, 31)

    def test_t_mode(self):
        self.assertEqual(TransmitMode(0), TransmitMode.TXA_TXB_RXA_RXB)
        self.assertEqual(TransmitMode(10), TransmitMode.TXA_TXB_RXA_RXB)

    def test_station_types(self):
        self.assertEqual(StationType(0), StationType.ALL)
        self.assertEqual(StationType(11), StationType.RESERVED)
        self.assertEqual(StationType(7), StationType.REGIONAL)
        self.assertEqual(StationType(16), StationType.ALL)

    def test_station_intervals(self):
        self.assertEqual(StationIntervals(0), StationIntervals.AUTONOMOUS_MODE)
        self.assertEqual(StationIntervals(11), StationIntervals.RESERVED)
        self.assertEqual(StationIntervals(7), StationIntervals.SECONDS_10)
        self.assertEqual(StationIntervals(12), StationIntervals.RESERVED)



class AISBaseException(Exception):
    """The base exception for all exceptions"""


class InvalidNMEAMessageException(AISBaseException):
    """Invalid NMEA Message"""
    pass


class InvalidNMEAChecksum(AISBaseException):
    """Invalid Checksum for the NMEA message"""


class UnknownMessageException(AISBaseException):
    """Message not supported yet"""
    pass


class MissingMultipartMessageException(AISBaseException):
    """Multipart message with missing parts provided"""


class TooManyMessagesException(AISBaseException):
    """Too many messages"""


class UnknownPartNoException(AISBaseException):
    """Unknown part number"""


class InvalidDataTypeException(AISBaseException):
    """An Unknown data type was passed to an encoding/decoding function"""


class NonPrintableCharacterException(AISBaseException):
    """A non printable ASCII character (0x20 (space) to 0x7e (~)) can not be decoded"""


class TagBlockNotInitializedException(Exception):
    """The TagBlock is not initialized"""


class MissingPayloadException(AISBaseException):
    """Valid NMEA Message without payload"""
    pass



import base64
import math
import typing
from collections import OrderedDict
from functools import partial, reduce
from operator import xor
from typing import Any, Generator, Hashable, TYPE_CHECKING, Union, Dict

from bitarray import bitarray

from pyais.constants import COUNTRY_MAPPING, SyncState
from pyais.exceptions import NonPrintableCharacterException

if TYPE_CHECKING:
    BaseDict = OrderedDict[Hashable, Any]
else:
    BaseDict = OrderedDict

from_bytes = partial(int.from_bytes, byteorder="big")
from_bytes_signed = partial(int.from_bytes, byteorder="big", signed=True)

T = typing.TypeVar('T')


def decode_into_bit_array(data: bytes, fill_bits: int = 0) -> bitarray:
    """
    Decodes a raw AIS message into a bitarray.
    :param data:        Raw AIS message in bytes
    :param fill_bits:   Number of trailing fill bits to be ignored
    :return:
    """
    bit_str = ''
    length = len(data)

    for i, c in enumerate(data):
        if not 0x20 <= c <= 0x7e:
            raise NonPrintableCharacterException(f"Non printable character: '{hex(c)}'")

        # Convert 8 bit binary to 6 bit binary
        c -= 0x30 if (c < 0x60) else 0x38
        c &= 0x3F

        if i == length - 1 and fill_bits:
            # The last part be shorter than 6 bits and contain fill bits
            c = c >> fill_bits
            bit_str += f'{c:b}'.zfill(6 - fill_bits)
        else:
            bit_str += f'{c:06b}'

    bit_arr = bitarray(bit_str)
    return bit_arr


def chunks(sequence: typing.Sequence[T], n: int) -> Generator[typing.Sequence[T], None, None]:
    """Yield successive n-sized chunks from sequence."""
    return (sequence[i:i + n] for i in range(0, len(sequence), n))


def decode_bin_as_ascii6(bit_arr: bitarray) -> str:
    """
    Decode binary data as 6 bit ASCII.
    :param bit_arr: array of bits
    :return: ASCII String
    """
    string: str = ""
    c: bitarray
    for c in chunks(bit_arr, 6):  # type:ignore
        n: int = from_bytes(c.tobytes()) >> 2

        # Last entry may not have 6 bits
        if len(c) != 6:
            n >> (6 - len(c))

        if n < 0x20:
            n += 0x40

        # Break if there is an @
        if n == 64:
            break

        string += chr(n)

    return string.strip()


def get_int(data: bitarray, ix_low: int, ix_high: int, signed: bool = False) -> int:
    """
    Cast a subarray of a bitarray into an integer.
    The bitarray module adds tailing zeros when calling tobytes(), if the bitarray is not a multiple of 8.
    So those need to be shifted away.
    :param data: some bitarray
    :param ix_low: the lower index of the sub-array
    :param ix_high: the upper index of the sub-array
    :param signed: True if the value should be interpreted as a signed integer
    :return: a normal integer (int)
    """
    shift: int = (8 - ((ix_high - ix_low) % 8)) % 8
    data = data[ix_low:ix_high]
    i: int = from_bytes_signed(data) if signed else from_bytes(data)
    return i >> shift


def checksum(sentence: bytes) -> int:
    """
    Compute the NMEA checksum for a payload.
    :param sentence: The sentence to compute the checksum for. MUST BE bytes.

    >>> checksum(b's:2573535,c:1671533231')
    8
    """
    checksum = reduce(xor, sentence)
    return checksum


def compute_checksum(msg: Union[str, bytes]) -> int:
    """
    Compute the checksum of a given message.
    This method takes the **whole** message including the leading `!`.

    >>> compute_checksum(b"!AIVDM,1,1,,B,15M67FC000G?ufbE`FepT@3n00Sa,0")
    91

    :param msg: message
    :return: int value of the checksum. Format as hex with `f'{checksum:02x}'`
    """
    if isinstance(msg, str):
        msg = msg.encode()

    msg = msg[1:].split(b'*', 1)[0]
    return reduce(xor, msg)


# https://gpsd.gitlab.io/gpsd/AIVDM.html#_aivdmaivdo_payload_armoring
PAYLOAD_ARMOR = {
    0: '0', 1: '1', 2: '2', 3: '3', 4: '4', 5: '5', 6: '6', 7: '7', 8: '8', 9: '9', 10: ':',
    11: ';', 12: '<', 13: '=', 14: '>', 15: '?', 16: '@', 17: 'A', 18: 'B', 19: 'C', 20: 'D',
    21: 'E', 22: 'F', 23: 'G', 24: 'H', 25: 'I', 26: 'J', 27: 'K', 28: 'L', 29: 'M', 30: 'N',
    31: 'O', 32: 'P', 33: 'Q', 34: 'R', 35: 'S', 36: 'T', 37: 'U', 38: 'V', 39: 'W', 40: '`',
    41: 'a', 42: 'b', 43: 'c', 44: 'd', 45: 'e', 46: 'f', 47: 'g', 48: 'h', 49: 'i', 50: 'j',
    51: 'k', 52: 'l', 53: 'm', 54: 'n', 55: 'o', 56: 'p', 57: 'q', 58: 'r', 59: 's', 60: 't',
    61: 'u', 62: 'v', 63: 'w'
}

# https://gpsd.gitlab.io/gpsd/AIVDM.html#_ais_payload_data_types
SIX_BIT_ENCODING = {
    '@': 0, 'A': 1, 'B': 2, 'C': 3, 'D': 4, 'E': 5, 'F': 6, 'G': 7, 'H': 8, 'I': 9, 'J': 10,
    'K': 11, 'L': 12, 'M': 13, 'N': 14, 'O': 15, 'P': 16, 'Q': 17, 'R': 18, 'S': 19, 'T': 20,
    'U': 21, 'V': 22, 'W': 23, 'X': 24, 'Y': 25, 'Z': 26, '[': 27, '\\': 28, ']': 29, '^': 30,
    '_': 31, ' ': 32, '!': 33, '"': 34, '#': 35, '$': 36, '%': 37, '&': 38, '\'': 39, '(': 40,
    ')': 41, '*': 42, '+': 43, ',': 44, '-': 45, '.': 46, '/': 47, '0': 48, '1': 49, '2': 50,
    '3': 51, '4': 52, '5': 53, '6': 54, '7': 55, '8': 56, '9': 57, ':': 58, ';': 59, '<': 60,
    '=': 61, '>': 62, '?': 63
}


def to_six_bit(char: str) -> str:
    """
    Encode a single character as six-bit bitstring.
    @param char: The character to encode
    @return: The six-bit representation as string
    """
    char = char.upper()
    try:
        encoding = SIX_BIT_ENCODING[char]
        return f"{encoding:06b}"
    except KeyError:
        raise ValueError(f"received char '{char}' that cant be encoded")


def encode_ascii_6(bits: bitarray) -> typing.Tuple[str, int]:
    """
    Transform the bitarray to an ASCII-encoded bit vector.
    Each character represents six bits of data.
    @param bits: The bitarray to convert to an ASCII-encoded bit vector.
    @return: ASCII-encoded bit vector and the number of fill bits required to pad the data payload to a 6 bit boundary.
    """
    out = ""
    chunk: bitarray
    padding = 0
    for chunk in chunks(bits, 6):  # type:ignore
        padding = 6 - len(chunk)
        num = from_bytes(chunk.tobytes()) >> 2
        armor = PAYLOAD_ARMOR[num]
        out += armor
    return out, padding


def int_to_bytes(val: typing.Union[int, bytes]) -> int:
    """
    Convert a bytes object to an integer. Byteorder is big.

    @param val: A bytes object to convert to an int. If the value is already an int, this is a NO-OP.
    @return: Integer representation of `val`
    """
    if isinstance(val, int):
        return val
    return int.from_bytes(val, 'big')


def bits2bytes(bits: typing.Union[str, bitarray]) -> bytes:
    """
    Convert a bitstring or a bitarray to bytes.
    >>> bits2bytes('00100110')
    b'&'
    """
    bits = bitarray(bits)
    return bits.tobytes()


def bytes2bits(in_bytes: bytes, default: typing.Optional[bitarray] = None) -> bitarray:
    """
    Convert a bytes object to a bitarray.

    @param  in_bytes :    The bytes to encode
    @param  default  :    A default value to return if `in_bytes` is *Falseish*

    >>> bytes2bits(b'&')
    bitarray('00100110')
    """
    if default is not None and not in_bytes:
        return default
    bits = bitarray(endian='big')
    bits.frombytes(in_bytes)
    return bits


def b64encode_str(val: bytes, encoding: str = 'utf-8') -> str:
    """BASE64 encoded a bytes string and returns the result as UTF-8 string"""
    return base64.b64encode(val).decode(encoding)


def coerce_val(val: typing.Any, d_type: typing.Type[T]) -> T:
    """Forces a given value in a given datatype"""
    if d_type == bytes and not isinstance(val, bytes):
        raise ValueError(f"Expected bytes, but got: {type(val)}")

    return d_type(val)  # type: ignore


def int_to_bin(val: typing.Union[int, bool], width: int, signed: bool = True) -> bitarray:
    """
    Convert an integer or boolean value to binary. If the value is too great to fit into
    `width` bits, the maximum possible number that still fits is used.

    @param val:     Any integer or boolean value.
    @param width:   The bit width. If less than width bits are required, leading zeros are added.
    @param signed:  Set to True/False if the value is signed or not.
    @return:        The binary representation of value with exactly width bits. Type is bitarray.
    """
    # Compute the total number of bytes required to hold up to `width` bits.
    n_bytes, mod = divmod(width, 8)
    if mod > 0:
        n_bytes += 1

    # If the value is too big, return a bitarray of all 1's
    mask = (1 << width) - 1
    if val >= mask:
        return bitarray('1' * width)

    bits = bitarray(endian='big')
    bits.frombytes(val.to_bytes(n_bytes, 'big', signed=signed))
    return bits[8 - mod if mod else 0:]


def str_to_bin(val: str, width: int, trailing_spaces: bool = False) -> bitarray:
    """
    Convert a string value to binary using six-bit ASCII encoding up to `width` chars.

    @param val:               The string to first convert to six-bit ASCII and then to binary.
    @param width:             The width of the full string
    @param trailing_spaces:   If the string has fewer characters than width, trailing '@' are added
    @return:        The binary representation of value with exactly width bits. Type is bitarray.
    """
    out = bitarray(endian='big')

    # Each char will be converted to a six-bit binary vector.
    # Therefore, the total number of chars is floor(WIDTH / 6).
    num_chars = int(width / 6)

    if trailing_spaces:
        # Add trailing '@' if the string is shorter than `width`
        for _ in range(num_chars - len(val)):
            val += "@"

    # Encode AT MOST width characters
    for char in val[:num_chars]:
        # Covert each char to six-bit ASCII vector
        txt = to_six_bit(char)
        out += bitarray(txt)

    return out


def chk_to_int(chk_str: bytes) -> typing.Tuple[int, int]:
    """
    Converts a checksum string to a tuple of (fillbits, checksum).
    >>> chk_to_int(b"0*1B")
    (0, 27)
    """
    if not len(chk_str):
        return 0, -1

    try:
        a, b = chk_str.split(b'*')
    except ValueError:
        return 0, -1

    try:
        fill_bits: int = int(a)
    except ValueError:
        fill_bits = 0

    try:
        checksum = int(b, 16)
    except (IndexError, ValueError):
        checksum = -1

    return fill_bits, checksum


SYNC_MASK = 0x03
TIMEOUT_MASK = 0x07
MSG_MASK = 0x3fff
SLOT_INCREMENT_MASK = 0x1fff


def get_sotdma_comm_state(radio: int) -> Dict[str, typing.Optional[int]]:
    """
    The SOTDMA communication state is structured as follows:
    +-------------------+----------------------+------------------------------------------------------------------------------------------------+
    | Parameter         |  Number of bits      |  Description                                                                                   |
    +-------------------+----------------------+------------------------------------------------------------------------------------------------+
    | Sync state        |  2                   |  0 UTC direct                                                                                  |
    |                   |                      |  1 UTC indirect                                                                                |
    |                   |                      |  2 Station is synchronized to a base station                                                   |
    |                   |                      |  3 Station is synchronized to another station based on the highest number of received stations |
    | Slot time-out     |  3                   |  Specifies frames remaining until a new slot is selected                                       |
    |                   |                      |  0 means that this was the last transmission in this slot                                      |
    |                   |                      |  1-7 means that 1 to 7 frames respectively are left until slot change                          |
    | Sub message       |  14                  |  14 The sub message depends on the current value in slot time-out                              |
    +-------------------+----------------------+------------------------------------------------------------------------------------------------+

    The slot time-out defines how to interpret the sub message:
    +-----------------+---------------------------------------------------------------------------+
    | Slot time-out   |  Description                                                              |
    +-----------------+---------------------------------------------------------------------------+
    | 3, 5, 7         |  Number of receiving stations (not own station) (between 0 and 16 383)    |
    | 2, 4, 6         |  Slot number Slot number used for this transmission (between 0 and 2 249) |
    | 1               |  UTC hour (bits 13 to 9) and minute (bits 8 to 2)                         |
    | 0               |  Next frame                                                               |
    +-----------------+---------------------------------------------------------------------------+

    You may refer to:
    - https://github.com/M0r13n/pyais/issues/17
    - https://www.itu.int/dms_pubrec/itu-r/rec/m/R-REC-M.1371-1-200108-S!!PDF-E.pdf
    - https://www.navcen.uscg.gov/?pageName=AISMessagesA#Sync
    """
    result = {
        'received_stations': None,
        'slot_number': None,
        'utc_hour': None,
        'utc_minute': None,
        'slot_offset': None,
        'slot_timeout': 0,
        'sync_state': 0,
    }

    sync_state = (radio >> 17) & SYNC_MASK  # First two (2) bits
    slot_timeout = (radio >> 14) & TIMEOUT_MASK  # Next three (3) bits
    sub_msg = radio & MSG_MASK  # Last 14 bits

    if slot_timeout == 0:
        result['slot_offset'] = sub_msg
    elif slot_timeout == 1:
        result['utc_hour'] = (sub_msg >> 9) & 0x1f
        result['utc_minute'] = (sub_msg >> 2) & 0x3f
    elif slot_timeout in (2, 4, 6):
        result['slot_number'] = sub_msg
    elif slot_timeout in (3, 5, 7):
        result['received_stations'] = sub_msg
    else:
        raise ValueError("Slot timeout can only be an integer between 0 and 7")

    result['sync_state'] = SyncState(sync_state)
    result['slot_timeout'] = slot_timeout
    return result


def get_itdma_comm_state(radio: int) -> Dict[str, typing.Optional[int]]:
    """
    +-----------------+------+--------------------------------------------------------------------------------+
    |    Parameter    | Bits |                                  Description                                   |
    +-----------------+------+--------------------------------------------------------------------------------+
    | Sync state      |   2  | 0 UTC direct                                                                   |
    |                 |      | 1 UTC indirec                                                                  |
    |                 |      | 2 Station is synchronized to a base station                                    |
    |                 |      | 3 Station is synchronized to another station                                   |
    | Slot increment  |  13  | Offset to next slot to be used, or zero (0) if no more transmissions           |
    | Number of slots |   3  | Number of consecutive slots to allocate. (0 = 1 slot, 1 = 2 slots,2 = 3 slots, |
    |                 |      | 3 = 4 slots, 4 = 5 slots)                                                      |
    | Keep flag       |   1  | Set to TRUE = 1 if the slot remains allocated for one additional frame         |
    +-----------------+------+--------------------------------------------------------------------------------+

    You may refer to:
    - https://github.com/M0r13n/pyais/issues/17
    - https://www.itu.int/dms_pubrec/itu-r/rec/m/R-REC-M.1371-1-200108-S!!PDF-E.pdf
    - https://www.navcen.uscg.gov/?pageName=AISMessagesA#Sync
    """

    sync_state = (radio >> 17) & SYNC_MASK  # First two (2) bits
    slot_increment = (radio >> 4) & SLOT_INCREMENT_MASK  # Next 13 bits
    num_slots = (radio >> 1) & TIMEOUT_MASK  # Next three (3) bits
    keep_flag = radio & 0x01  # Last bit

    return {
        'keep_flag': keep_flag,
        'sync_state': sync_state,
        'slot_increment': slot_increment,
        'num_slots': num_slots,
        'keep_flag': keep_flag,
    }


def get_first_three_digits(num: int) -> int:
    if num < 1000:
        return num
    digits = int(math.log10(num)) + 1
    return int(num // (10**(digits - 3)))


def get_country(mmsi: int) -> typing.Tuple[str, str]:
    return COUNTRY_MAPPING.get(get_first_three_digits(mmsi), ('NA', 'Unknown'))



import typing
import pathlib
import queue
from abc import ABC, abstractmethod
from socket import AF_INET, SOCK_DGRAM, SOCK_STREAM, socket
from typing import BinaryIO, Generator, Generic, Iterable, List, TypeVar, cast

from pyais.exceptions import InvalidNMEAMessageException, NonPrintableCharacterException, UnknownMessageException
from pyais.messages import AISSentence, GatehouseSentence, NMEAMessage, NMEASentence, NMEASentenceFactory

T = TypeVar("T")
F = TypeVar("F", BinaryIO, socket, None)
DOLLAR_SIGN = ord("$")
EXCLAMATION_POINT = ord("!")
BACKSLASH = ord("\\")


def should_parse(byte_str: bytes) -> bool:
    """Return True if a given byte string seems to be NMEA message.
    This method does **NOT** validate the message, but uses a heuristic
    approach to check (or guess) if byte string is a valid nmea_message.
    """
    # The byte sequence is not empty and starts with a $ or a ! or \
    return len(byte_str) > 0 and byte_str[0] in (DOLLAR_SIGN, EXCLAMATION_POINT, BACKSLASH)


class PreprocessorProtocol(typing.Protocol):
    """
    Protocol for preprocessing lines of bytes into NMEA0183 format.

    This protocol must be implemented to support various message formats in the pyais library.
    Implementing classes should provide a `process` method that transforms input lines of bytes
    into NMEA0183 compliant messages.

    Methods:
    --------
    process(line: bytes) -> bytes:
        Processes an input line of bytes and returns it in NMEA0183 format.
    """

    def process(self, line: bytes) -> bytes:
        """
        Process an input line of bytes and return it in NMEA0183 format.

        Parameters:
        -----------
        line : bytes
            The input line of bytes to be processed.

        Returns:
        --------
        bytes
            The processed line in bytes, conforming to the NMEA0183 format.
        """
        pass


class TagBlockQueue(queue.Queue):  # type: ignore

    def __init__(self, maxsize: int = 0) -> None:
        self.groups: typing.Dict[str, typing.Dict[str, object]] = {}
        super().__init__(maxsize)

    def put_sentence(self, sentence: NMEASentence) -> None:
        if not sentence.tag_block:
            # No NMEA 4.10 tag block
            # Returns a single line immediately.
            super().put([sentence,])
            return

        tb = sentence.tag_block
        tb.init()

        if not tb.group:
            # No NMEA 4.10 tag block 'g'.
            super().put([sentence,])
            return

        if int(tb.group.sentence_tot) == 1:
            # Group of a single sentence
            super().put([sentence,])
            return

        if int(tb.group.sentence_num) == 1:
            # The first sentence
            self.groups[tb.group.group_id] = {'sentence_tot': tb.group.sentence_tot, 'sentences': [sentence,]}
            return

        if tb.group.group_id not in self.groups:
            # Unknown group. First sentence of group is missing.
            return

        self.groups[tb.group.group_id]['sentences'].append(sentence)  # type: ignore

        # is_last_sentence = int(tb.group.sentence_num) != int(self.groups[tb.group.group_id]['sentence_tot'])
        group_is_complete = int(tb.group.sentence_tot) != len(self.groups[tb.group.group_id]['sentences'])  # type: ignore

        if group_is_complete:
            return

        # All sentences belonging to this group were received.
        super().put(self.groups[tb.group.group_id]['sentences'])
        del self.groups[tb.group.group_id]


class AssembleMessages(ABC):
    """
    Base class that assembles multiline messages.
    Offers a iterator like interface.
    """

    def __init__(self, tbq: typing.Optional[TagBlockQueue] = None) -> None:
        self.wrapper_msg: typing.Optional[GatehouseSentence] = None
        self.tbq: typing.Optional[TagBlockQueue] = tbq

    def __enter__(self) -> "AssembleMessages":
        # Enables use of with statement
        return self

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        return None

    def __iter__(self) -> Generator[NMEAMessage, None, None]:
        return self._assemble_messages()

    def __next__(self) -> NMEAMessage:
        """Returns the next decoded NMEA message."""
        return next(iter(self))

    def __set_last_wrapper_msg(self, wrapper_msg: GatehouseSentence) -> None:
        self.wrapper_msg = wrapper_msg

    def __get_last_wrapper_msg(self) -> typing.Optional[GatehouseSentence]:
        wrapper_msg = self.wrapper_msg
        self.wrapper_msg = None
        return wrapper_msg

    def __insert_wrapper_msg(self, msg: AISSentence) -> AISSentence:
        wrapper_msg = self.__get_last_wrapper_msg()
        if wrapper_msg:
            msg.wrapper_msg = wrapper_msg
        return msg

    def __add_to_tbq(self, sentence: NMEASentence) -> None:
        if not self.tbq:
            # Tag Block Queue not defined. Do nothing.
            return
        self.tbq.put_sentence(sentence)

    def _assemble_messages(self) -> Generator[NMEAMessage, None, None]:
        buffer: typing.Dict[typing.Tuple[int, str], typing.List[typing.Optional[NMEAMessage]]] = {}
        messages = self._iter_messages()
        msg: AISSentence

        for line in messages:
            try:
                sentence = NMEASentenceFactory.produce(line)
                self.__add_to_tbq(sentence)
                if sentence.TYPE == GatehouseSentence.TYPE:
                    sentence = cast(GatehouseSentence, sentence)
                    self.__set_last_wrapper_msg(sentence)
                    continue
            except (InvalidNMEAMessageException, NonPrintableCharacterException, UnknownMessageException):
                # Be gentle and just skip invalid messages
                continue

            if not sentence.TYPE == AISSentence.TYPE:
                continue
            msg = typing.cast(AISSentence, sentence)

            if msg.is_single:
                yield self.__insert_wrapper_msg(msg)
            else:
                # Instead of None use -1 as a seq_id
                seq_id = msg.seq_id
                if seq_id is None:
                    seq_id = -1

                # seq_id and channel make a unique stream
                slot = (seq_id, msg.channel)

                if slot not in buffer:
                    # Create a new array in the buffer that has enough space for all fragments
                    buffer[slot] = [None, ] * max(msg.fragment_count, 0xff)

                buffer[slot][msg.frag_num - 1] = msg
                msg_parts = buffer[slot][0:msg.fragment_count]

                # Check if all fragments are found
                not_none_parts = [m for m in msg_parts if m is not None]
                if len(not_none_parts) == msg.fragment_count:
                    msg = NMEAMessage.assemble_from_iterable(not_none_parts)
                    yield self.__insert_wrapper_msg(msg)
                    del buffer[slot]

    @abstractmethod
    def _iter_messages(self) -> Generator[bytes, None, None]:
        raise NotImplementedError("Implement me!")


class IterMessages(AssembleMessages):

    def __init__(self, messages: Iterable[bytes], tbq: typing.Optional[TagBlockQueue] = None):
        super().__init__(tbq=tbq)
        # If the user passes a single byte string make it into a list
        if isinstance(messages, bytes):
            messages = [messages, ]
        self.messages: Iterable[bytes] = messages

    @classmethod
    def from_strings(cls, messages: Iterable[str], ignore_encoding_errors: bool = False,
                     encoding: str = "utf-8") -> "IterMessages":
        # If the users passes a single message as string, make it a list
        if isinstance(messages, str):
            messages = [messages, ]

        encoded: List[bytes] = []
        for message in messages:
            try:
                encoded.append(message.encode(encoding))
            except UnicodeEncodeError as e:
                if ignore_encoding_errors:
                    # Just skip and carry on
                    continue
                raise e

        return IterMessages(encoded)

    def _iter_messages(self) -> Generator[bytes, None, None]:
        # Transform self.messages into a generator
        yield from (message for message in self.messages)


class Stream(AssembleMessages, Generic[F], ABC):

    def __init__(
        self, fobj: F,
        preprocessor: typing.Optional[PreprocessorProtocol] = None,
        tbq: typing.Optional[TagBlockQueue] = None
    ) -> None:
        """
        Create a new Stream-like object.
        @param fobj: A file-like or socket object.
        @param preprocessor: An optional preprocessor
        """
        super().__init__(tbq=tbq)
        self._fobj: F = fobj
        self.preprocessor = preprocessor

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        self.close()

    def _iter_messages(self) -> Generator[bytes, None, None]:
        # Do not parse lines, that are obviously not NMEA messages
        for line in self.read():
            if len(line) <= 10:
                continue
            if self.preprocessor is not None:
                line = self.preprocessor.process(line)
            if should_parse(line):
                yield line

    def close(self) -> None:
        if self._fobj is not None:
            self._fobj.close()

    @abstractmethod
    def read(self) -> Generator[bytes, None, None]:
        raise NotImplementedError()


class BinaryIOStream(Stream[BinaryIO]):
    """Read messages from a file-like object"""

    def __init__(
        self,
        file: BinaryIO,
        preprocessor: typing.Optional[PreprocessorProtocol] = None,
        tbq: typing.Optional[TagBlockQueue] = None
    ) -> None:
        super().__init__(file, preprocessor=preprocessor, tbq=tbq)

    def read(self) -> Generator[bytes, None, None]:
        yield from self._fobj


class FileReaderStream(BinaryIOStream):
    """
    Read NMEA messages from file
    """

    def __init__(
        self,
        filename: typing.Union[str, pathlib.Path],
        mode: str = "rb",
        preprocessor: typing.Optional[PreprocessorProtocol] = None,
        tbq: typing.Optional[TagBlockQueue] = None
    ) -> None:
        self.filename: typing.Union[str, pathlib.Path] = filename
        self.mode: str = mode
        # Try to open file
        try:
            file = open(self.filename, mode=self.mode)
            file = cast(BinaryIO, file)
        except Exception as e:
            raise FileNotFoundError(f"Could not open file {self.filename}") from e
        super().__init__(file, preprocessor=preprocessor, tbq=tbq)


class ByteStream(Stream[None]):
    """
    Takes a iterable that contains ais messages as bytes and assembles them.
    """

    def __init__(
        self,
        iterable: Iterable[bytes],
        preprocessor: typing.Optional[PreprocessorProtocol] = None,
        tbq: typing.Optional[TagBlockQueue] = None
    ) -> None:
        self.iterable: Iterable[bytes] = iterable
        super().__init__(None, preprocessor=preprocessor, tbq=tbq)

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        return

    def read(self) -> Generator[bytes, None, None]:
        yield from self.iterable


class SocketStream(Stream[socket]):
    BUF_SIZE = 4096

    def recv(self) -> bytes:
        return self._fobj.recv(self.BUF_SIZE)

    def read(self) -> Generator[bytes, None, None]:
        partial: bytes = b''
        while True:
            body = self.recv()

            # Server closed connection
            if not body:
                return None

            lines = body.splitlines(keepends=True)

            # last incomplete line
            line = partial + lines[0]
            if line:
                yield line
            partial = b''

            if lines[-1].endswith(b'\n'):
                # all lines are complete
                yield from lines[1:]
            else:
                # the last line was only partially received
                yield from lines[1:-1]
                partial = lines[-1]


class UDPReceiver(SocketStream):

    def __init__(
        self,
        host: str,
        port: int,
        preprocessor: typing.Optional[PreprocessorProtocol] = None,
        tbq: typing.Optional[TagBlockQueue] = None
    ) -> None:
        sock: socket = socket(AF_INET, SOCK_DGRAM)
        sock.bind((host, port))
        super().__init__(sock, preprocessor=preprocessor, tbq=tbq)

    def recv(self) -> bytes:
        return self._fobj.recvfrom(self.BUF_SIZE)[0]


class TCPConnection(SocketStream):
    """
     Read AIS data from a remote TCP server
     https://en.wikipedia.org/wiki/NMEA_0183
     """

    def recv(self) -> bytes:
        return self._fobj.recv(self.BUF_SIZE)

    def __init__(
        self,
        host: str,
        port: int = 80,
        preprocessor: typing.Optional[PreprocessorProtocol] = None,
        tbq: typing.Optional[TagBlockQueue] = None
    ) -> None:
        sock: socket = socket(AF_INET, SOCK_STREAM)
        try:
            sock.connect((host, port))
        except ConnectionRefusedError as e:
            sock.close()
            raise ConnectionRefusedError(f"Failed to connect to {host}:{port}") from e
        super().__init__(sock, preprocessor=preprocessor, tbq=tbq)
